__doc__='''Poll Object'''
__version__='0.7.1'

from Globals import HTMLFile, MessageDialog, package_home
from Globals import Persistent
from Acquisition import Implicit
from AccessControl.Role import RoleManager
import OFS.SimpleItem 
from OFS.DTMLMethod import DTMLMethod
from OFS.Folder import Folder
import re, os, string
import cStringIO

no_pil=0
try: import PIL
except: no_pil=1

addItemForm=HTMLFile('itemAdd',globals())
def addItem(self, id, title='', questions = [], REQUEST=None):
    """Add an item to a folder

    The argument 'self' will be bound to a folder.

    The arguments, 'id' and 'title' are minimal, and most applications
    will have additional needed arguments.
    """
    self._setObject(id, Poll(id, title))
    if REQUEST is not None: return self.manage_main(self,REQUEST)

class Poll(
    Folder,
    RoleManager
    ):
    """A Poll Object """

    meta_type='Poll'
    id       ='Poll'
    title = 'Poll Object'
    if no_pil: fancy_graphics=0
    else: fancy_graphics=1

    isPrincipiaFolderish = 1

    manage_options=(
        {'label':'Properties', 'action':'manage_main'},
        {'label':'Documents',  'action':'manage_documents'},
        {'label':'View',       'action':''},
        {'label':'Security',   'action':'manage_access'},
        {'label':'Voters',     'action':'manage_voters'},
        )

    __ac_permissions__=(
        ('View management screens', ('manage_tabs','manage_main')),
        ('Change permissions',      ('manage_access',)           ),
        ('Change Poll',             ('manage_edit','manage_addQuestion',
                                     'manage_documents',
                                     'manage_voters','deleteVoters')),
        ('View',                    ('',)                        ),
        ('Vote',                    ('processVotes',)            ),
        ('Change permissions',      ('manage_access',)           ),
        )


    def __init__(self, id, title=''):
        self.id=id
        self.title=title
        self.questionNumber = 1
        self.amPoll = 1
        self.open = 1
        self.anonymous = 1
        self.voted = {}

        prefix = package_home(globals())

        # Add a Document used to view previous entries
        fileName = os.path.join(prefix,'index.dtml')
        index2 = open(fileName,'r').read()
        i = DTMLMethod(index2,__name__='index_html')
        i.title=title
        self._setObject('index_html',i)

        fileName = os.path.join(prefix,'results.dtml')
        index2 = open(fileName,'r').read()
        i = DTMLMethod(index2,__name__='results')
        self._setObject('results',i)

    manage_main=HTMLFile('pollEdit', globals())
    manage_documents=HTMLFile('pollDocuments', globals())
    manage_voters=HTMLFile('pollVoters', globals())
    questionAdd = HTMLFile('questionAdd', globals())

    def tpValues(self):
        """Returns a list of the folders sub-folders, used by tree tag. (Stolen from Folder)"""
        r=[]
        for id in self._objects:
            o=getattr(self, id['id'])
            try:
                if o.isPrincipiaFolderish: r.append(o)
            except: pass
        return r

    def processVotes(self, responses=[], REQUEST=None):
        """Gather form data and increment the appropriate responses"""
        if not self.open:
            if REQUEST is not None:
                return MessageDialog(title='Error',
                                     message='This poll is not open.',
                                     action ='.'
                                     )
            else:
                return

        if not self.anonymous and REQUEST is not None and REQUEST.has_key('AUTHENTICATED_USER'):
            if self.voted.has_key(REQUEST['AUTHENTICATED_USER'].getUserName()):
                raise 'Vote error', 'You have already voted for this poll.'
            else:
                voted=self.voted
                voted[REQUEST['AUTHENTICATED_USER'].getUserName()] = ''
                self.voted=voted

        myRe = re.compile('(question.*)(response.*)')

        n_answered = 0
        for response in responses:
            if response != '---No selection---':
                (questionId, responseId) = myRe.match(response).groups()
                question = getattr(self, questionId)
                response = getattr(question, responseId)
                response.increment()
                question.totalVotes = question.totalVotes + 1
                n_answered = n_answered + 1
            else:
                raise 'Vote error', 'You must answer all questions.'

        if n_answered != len(self.objectIds(['Poll Question'])):
            raise 'Vote error', 'You must answer all questions.'
            
        return MessageDialog(title='Edited',
                             message='Votes accepted.',
                             action ='./results'
                             )


    def makeQuestionId(self):
        questionNumber = 1
        questionId = ''
        currentIds = self.objectIds(['Poll Question'])

        while 1:
            newId = "question%s" % questionNumber
            if not newId in currentIds:
                break
            questionNumber = questionNumber + 1
        return newId

    def manage_addQuestion(self, question = '', responses = []):
        """Add a question"""
        newId = self.makeQuestionId()
        newQuestion = Question(newId, question, responses,'bla')
        self._setObject(newId, newQuestion)
        return MessageDialog(title='Edited',
                             message='Entry has been edited.',
                             action ='./manage_main',
                             )

    def manage_edit(self, title, open=0, anonymous=0, REQUEST=None):
        """Edit the Poll"""
        self.title=title
        self.open=open
        self.anonymous=anonymous
        if REQUEST is not None:
            return MessageDialog(
                title='Edited',
                message='<strong>%s</strong> has been edited.' % self.id,
                action ='./manage_main',
                )

    def deleteVoters(self, REQUEST=None):
        """Delete all voters"""
        self.voted = {}
        if REQUEST is not None:
            return MessageDialog(
                title='Voters deleted',
                message='<strong>%s</strong> has been edited.' % self.id,
                action ='./manage_voters',
                )

class Question(
    Folder
    ):
    """A Question belonging to a Poll"""

    isPrincipiaFolderish = 1

    meta_type='Poll Question'
    id       ='Poll Question'
    icon = 'misc_/Poll/q'

    title = 'Question Object'

    # graph types are registered later
    graphTypes = {}

    index_html=HTMLFile('questionView' , globals())
    manage_main=HTMLFile('questionEdit',globals())
    legendTemplate=HTMLFile('legend', globals())

    # Specify definitions for tabs:
    manage_options=(
        {'label':'Properties', 'action':'manage_main'},
        {'label':'View',       'action':''},
        {'label':'Security',   'action':'manage_access'},
        )

    def __init__(self, id, question, responses, size = 500):
        self.id = id
        self.questionId = id
        self.question = question
        self.totalVotes = 0.
        self.amPoll = 0
        # for the time being pieChart is the only one
        self.graphType = 'Pie chart'

        for response in responses:
            if len(response):
                newId = self.makeResponseId()
                newResponse = Response( newId, response)
                self._setObject(newId, newResponse)
        self.size = size

    def manage_clearAllVotes(self, REQUEST= None):
        """Delete all votes. i.e. Reset the Poll"""
        for response in self.objectValues(['Poll Response']):
            response.votes = 0
        self.totalVotes = 0
        if not REQUEST is None:
            return MessageDialog(title='Cleared',
                                 message='All votes have been deleted',
                                 action ='./manage_main',
                                 )
    def makeResponseId(self):
        responseNumber = 1
        responseId = ''
        currentIds = self.objectIds(['Poll Response'])

        while 1:
            newId = "response%s" % responseNumber
            if not newId in currentIds:
                break
            responseNumber = responseNumber + 1
        return newId

    def manage_addResponse(self, newResponse ='', REQUEST=None):
        """Add a response"""
        newId = self.makeResponseId()
        newResponse = Response(newId, newResponse)
        self._setObject(newId, newResponse)

        return MessageDialog(title='Added',
                             message='Response has been added.',
                             action ='./manage_main',
                             )

    def manage_delResponses(self, ids=[], REQUEST=None):
        """Delete a set of responses specified by ids"""
        for id in ids:
            try:
                response = getattr(self, id)
                self.totalVotes = self.totalVotes - response.votes
            except:
                pass
        return self.manage_delObjects(ids, REQUEST=REQUEST)
        
    def manage_updateVotes(self, votes=[], realIds = [], responses = [], REQUEST=None):
        """Change votes... useful for vote-riggers ;)"""

        if not self.objectIds(['Poll Response']) == realIds:
            return MessageDialog(
                title='Form Out of Date',
                message='Form is out of date - Edit again', 
                action ='./manage_main',
                )

        total = 0
        for i in range(len(realIds)):
            vote = string.atof(votes[i])
            id   = realIds[i]
            response = getattr(self, id)
            total = total + vote
            response.votes = vote
            response.response = responses[i]

        self.totalVotes = total
        
        if REQUEST is not None:
            return MessageDialog(
                title='Updated',
                message='Question Changed', 
                action ='./manage_main',
                )
        
    def manage_clearVotes(self, ids=[], REQUEST=None):
        """Clear the votes for the response responses specified by ids"""
        for id in ids:
            try:
                response = getattr(self, id)
                self.totalVotes = self.totalVotes - response.votes
                response.votes = 0
            except:
                pass
        if REQUEST is not None:
            return MessageDialog(
                title='Cleared',
                message='Votes for selected responses have been cleared.', 
                action ='./manage_main',
                )

    def tpValues(self):
        """Returns a list of the folders sub-folders, used by tree tag."""
        r=[]
        for id in self._objects:
            o=getattr(self, id['id'])
            try:
                if o.isPrincipiaFolderish: r.append(o)
            except: pass
        return r

    def manage_edit(self, question, votes = [], realIds = [], responses = [], REQUEST=None):
        """Edit the Question"""
        currentObjects =self.objectIds(['Poll Response'])

        realIds.sort()
        currentObjects.sort()

        self.question = question
        if not currentObjects == realIds:
            return MessageDialog(
                title='Form Out of Date',
                message='Form is out of date - Edit again', 
                action ='./manage_main',
                )

        total = 0
        for i in range(len(realIds)):
            vote = string.atof(votes[i])
            id   = realIds[i]
            response = getattr(self, id)
            total = total + vote
            response.votes = vote
            response.response = responses[i]

        self.totalVotes = total
        
        if REQUEST is not None:
            return MessageDialog(
                title='CUpdates',
                message='Votes Updated.', 
                action ='./manage_main',
                )

    def getColors(self):
        red = (255,0,0)
        green = (0,255,255)
        blue = (0,0,255)
        cyan = (0,255,255)
        green = (0,255,0)
        yellow = (255,255,0)
        orange = (255,165,0)
        magenta = (255,0,255)


        # Feel free to change these to whatever is 'in' this summer/
        colors = [ red, green, blue, cyan, yellow, orange, magenta]
        return colors

    def legend(self):
        return self.drawLegend(amPoll = 0)

    def pollLegend(self):
        return self.drawLegend(amPoll = 1)

    def drawLegend(self, amPoll = 1):
        """Render an HTML legend for the piechart"""
        color = 0
        colors = self.getColors()
        colorTable = []
        
        for response in self.objectValues(['Poll Response']):
            if response.votes:
                voteString = "%s (%.2f%%)" % (response.votes, 100*response.votes/self.totalVotes)
            else:
                voteString = "%s" % (response.votes, )
            
            currentColor = colors[color]
            colorTable.append(  {'response'   : response.response,
                                 'colorCode'  :"#%02X%02X%02X" % currentColor,
                                 'voteString' : voteString,
                                 'questionId' : self.questionId,
                                 'responseId' : response.responseId,
                                 'amPoll'     : amPoll })
            color = color + 1
            if color == len(colors):
                color = 0
            
        return self.legendTemplate(self, table = colorTable)

    def drawGraph(self, size=500, RESPONSE=None):
        if no_pil:
            drawNoGraph(self, size, RESPONSE)
        else:
            apply(self.graphTypes[self.graphType], (self, size, RESPONSE) )

    def drawPieChart(self, size=500, RESPONSE=None):
        """Draw a pie chart. The sole purpose of counter is to keep IMG refererences unique"""
        import sys
        im = PIL.Image.new('RGB', (size,size))
        draw = PIL.ImageDraw.ImageDraw(im)
        bgcolor = (255, 255, 255)
        
        draw.rectangle([0, 0, size, size], fill=bgcolor, outline=bgcolor)
        theta = 0.0
        color = 0
        colors = self.getColors()

        for response in self.objectValues(['Poll Response']):
            
            if response.votes:    
                newTheta = theta + (response.votes/self.totalVotes) * 360.0
                
                draw.setink(colors[color])
                draw.setfill(1)
                draw.pieslice( (0,0,size, size), theta, newTheta)
                theta = newTheta
            
            color = color + 1
            if color == len(colors):
                color = 0
                
        del draw

        outFile = cStringIO.StringIO()
        im.save(outFile, 'GIF')

        outFile.seek(0)
        RESPONSE.setHeader("Content-type", "image/gif")
        RESPONSE.write(outFile.read())
        return ''

    graphTypes['Pie chart']=drawPieChart

    def drawNothing(self, size=500, RESPONSE=None):
        """Draw nothing at all"""
        return "\n\n"

    graphTypes['Nothing']=drawNothing

    def drawBarGraph(self, size=500, RESPONSE=None):
        """Draw a bar graph. Not implemented yet."""
        return "\n\n"

    graphTypes['Bar graph']=drawBarGraph

class Response(
    Implicit,
    RoleManager,
    Persistent,
    OFS.SimpleItem.Item
    ):
    meta_type='Poll Response'
    id       ='Poll Response'
    icon = 'misc_/Poll/a'

    title = 'Response Object'

    index_html=HTMLFile('responseView', globals())
    manage_main=HTMLFile('responseEdit' ,globals())

    # Specify definitions for tabs:
    manage_options=(
        {'label':'Properties', 'action':'manage_main'},
        {'label':'View',       'action':''},
        {'label':'Security',   'action':'manage_access'},
        )

    def __init__(self, id, response):
        self.id = id
        self.responseId = id
        self.response = response
        self.votes    = 0

    def setVotes(self, votes):
        self.votes = votes

    def increment(self):
        self.votes = self.votes + 1

    def manage_edit(self, response, votes, REQUEST=None):
        """Edit the Response"""
        
        self.response = response

        difference = votes - self.votes
        self.aq_parent.totalVotes = self.aq_parent.totalVotes + difference
        self.setVotes(votes)

        if REQUEST is not None:
            return MessageDialog(
                title='Edited',
                message='<strong>%s</strong> has been edited.' % self.id,
                action ='./manage_main',
                )
